;; -*- lisp -*-

(in-package :it.bese.ucw)

;;;; ** STANDARD-SESSION

(defmethod print-object ((s standard-session) stream)
  (print-unreadable-object (s stream :type t :identity t)
    (format stream "~D ~S" (hash-table-count (session.frames s)) (session.current-frame s))))

(defvar *standard-session-longevity* (* 60 30)
  "Amout of inactivity, in seconds, allowed before a
  standard-session object is expired.")

(defmethod session.value (key (session standard-session) &optional default)
  "Lookup the value associated with KEY in the session."
  (gethash key (session.object-pool session) default))

(defmethod (setf session.value) (value key (session standard-session) &optional default)
  (setf (gethash key (session.object-pool session) default) value))

(defmethod expiredp ((session standard-session))
  "Returns T if SESSION's last-access is more than
*standard-session-longevity* seconds ago."
  (< (session.last-access session)
     (- (get-universal-time) *standard-session-longevity*)))

(defmethod expire-session ((session standard-session))
  nil)

(defmethod find-frame ((s standard-session) (frame-id string))
  "Returns the frame whose id is FRAME-ID."
  (flet ((queue-first (q)
           (unless (cdr q)
             (setf (cdr q) (nreverse (car q))
                   (car q) nil))
           (car (cdr q))))
    (or (gethash frame-id (session.frames s))
        (if-bind frame-id (queue-first (session.frame-queue s))
          (gethash frame-id (session.frames s))))))

(defmethod make-new-frame ((session standard-session))
  "Inserts a new frame in the session stack of SESSION. Returns
  the new frame. This new frame will have the same component and
  backtracks as the current frame in SESSION. Sets SESSION's
  current-frame to the newly created frame."
  (let* ((id (loop
                for id = (random-string +frame-id-length+)
                while (gethash id (session.frames session))
                finally (return id)))
         (new-frame (make-next-frame (session.current-frame session) id)))
    (flet ((enqueue-frame ()
             (push id (car (session.frame-queue session)))
             (setf (gethash id (session.frames session)) new-frame))
           (dequeue-frame ()
             (with-slots (frame-queue) session
               (unless (cdr frame-queue)
                 (setf (cdr frame-queue) (nreverse (car frame-queue))
                       (car frame-queue) nil))
               (when (cdr frame-queue)
                 (remhash (pop (cdr frame-queue)) (session.frames session))))))
      
      (when (>= (hash-table-count (session.frames session))
                +session-backtracking-max-depth+)
        (dequeue-frame))
      
      (enqueue-frame)
      (setf (session.current-frame session) new-frame))))

(defconstant +epoch-start+ (encode-universal-time 0 0 0 1 1 1970))

(defmethod service :before ((session standard-session) (context request-context))
  (setf (session.last-access session) (get-universal-time)))

(defmethod service ((session standard-session) (context request-context))
  "Service a request for a session.

If we find a frame-id in CONTEXT then we either service the
associated request or, should this not exist, hope the request
has an entry-point and call that (errors then will come out when
servicing the frame).

If CONTEXT doesn't provide a frame-id we just assume this is an
entry-point request, create a new frame and service it."
  (if-bind frame-id (find-frame-id context)
      (if-bind frame (find-frame session frame-id)
          (progn
            (setf (session.current-frame session) frame)
            (service frame context))
          (progn
            (ucw.rerl.session.warn "Inexistent frame-id: ~S." frame-id)
            (service (make-new-frame session) context)))
      (service (make-new-frame session) context)))

;; Copyright (c) 2003-2005 Edward Marco Baringer
;; All rights reserved. 
;; 
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are
;; met:
;; 
;;  - Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;; 
;;  - Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;; 
;;  - Neither the name of Edward Marco Baringer, nor BESE, nor the names
;;    of its contributors may be used to endorse or promote products
;;    derived from this software without specific prior written permission.
;; 
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
;; "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
;; LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
;; A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
;; OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
;; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
;; LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
;; DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
;; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
;; (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
